from typing import Dict, List, Union
from enum import EnumMeta
import inspect
import inflection
from dataclasses import is_dataclass


# dataclass field(init=False)로 지정되어 있는 변수들은
# constructor에 넣어서 생성 할 수 없다.
# 따라서, 생성자에 넘겨줄 dictionary와 setattr로 설정할 변수를 나누어 설정한다.
def from_dict(dataclass, dict: Dict):
    # Allow None value for required false value
    if dict is None:
        return None

    sig = inspect.signature(dataclass.__init__)

    init_dict = {}
    init_false_dict = {}
    for key, val in dict.items():
        # camel case to snake case
        key = inflection.underscore(key)

        if key in sig.parameters.keys():
            key_class = sig.parameters[key].annotation
            if is_dataclass(key_class):
                init_dict[key] = from_dict(key_class, val)
            # regard as typing hint
            elif hasattr(key_class, "__origin__") and (
                key_class.__origin__ == list or key_class.__origin__ == List
            ):
                if is_dataclass(key_class.__args__[0]):
                    init_dict[key] = [
                        from_dict(key_class.__args__[0], elem) for elem in val
                    ]
                # handle List[Union[XXX, YYY]] type
                elif hasattr(key_class.__args__[0], '__origin__') and (key_class.__args__[0].__origin__ == Union):
                    def union_mapping(union_classes, value):
                        for union_class in union_classes.__args__:
                            if value is None or (
                                    hasattr(union_class, "__annotations__") and
                                    union_class.__annotations__.keys() == {
                                        inflection.underscore(v_k) for v_k in value.keys()}
                            ):
                                try:
                                    return from_dict(union_class, value)
                                except AssertionError:
                                    continue
                        return value
                    init_dict[key] = [union_mapping(key_class.__args__[0], v) for v in val]
                else:
                    init_dict[key] = val
            # handle Union[XXX, YYY] type
            elif (
                    hasattr(key_class, '__origin__') and key_class.__origin__ == Union
            ):
                def union_mapping(union_classes, value):
                    for union_class in union_classes.__args__:
                        if is_dataclass(union_class) and (value is None or (
                                hasattr(union_class, "__annotations__") and
                                union_class.__annotations__.keys() == {inflection.underscore(v_k) for v_k in value.keys()}
                        )):
                            try:
                                return from_dict(union_class, value)
                            except AssertionError:
                                continue
                    return value
                if type(val) == list:
                    init_dict[key] = [union_mapping(key_class, v) for v in val]
                else:
                    init_dict[key] = union_mapping(key_class, val)
            elif isinstance(key_class, EnumMeta):
                init_dict[key] = key_class(val)
            else:
                assert isinstance(val, dataclass.__annotations__[key]), \
                    f'{key} type must be {dataclass.__annotations__[key]} (type: {type(val)}, value: {val}))'
                init_dict[key] = val
        else:
            init_false_dict[key] = val

    instance = dataclass(**init_dict)
    for key, val in init_false_dict.items():
        setattr(instance, key, val)
    return instance


def to_dict(instance):
    result = {}
    for k, v in instance.__dict__.items():
        if is_dataclass(v):
            result[inflection.camelize(k, False)] = to_dict(v)
        elif isinstance(v, list):
            result[inflection.camelize(k, False)] = []
            for v_elem in v:
                if is_dataclass(v_elem):
                    result[inflection.camelize(k, False)].append(to_dict(v_elem))
                else:
                    result[inflection.camelize(k, False)].append(v_elem)
        else:
            result[inflection.camelize(k, False)] = v
    return result



from .experiment_predefinedai_dto import ExperimentPredefinedaiDto
from .experiment_automl_dto import ExperimentAutomlDto
from .experiment_mllab_dto import ExperimentMllabDto
from .experiment_data_object_info import ExperimentDataObjectInfo

__all__ = ["ExperimentPredefinedaiDto", "ExperimentAutomlDto", "ExperimentMllabDto", "ExperimentDataObjectInfo"]



from dataclasses import dataclass
from enum import Enum
import abc
from typing import List, Union


class StrEnum(str, Enum):
    def __repr__(self):
        return str(self.name)

    def __str__(self):
        return str(self.name)


class PredefinedAIArgType(StrEnum):
    String = 'String'
    Integer = 'Integer'
    Float = 'Float'
    Categorical = 'Categorical'
    MultipleOf2 = 'MultipleOf2'
    Bool = 'Bool'


class PredefinedAIArg:
    @abc.abstractmethod
    def get_key(self):
        raise NotImplementedError

    @abc.abstractmethod
    def get_value(self):
        raise NotImplementedError


class ArgGroup(StrEnum):
    Arg = 'Arg'
    AlgorithmHyperparameter = 'AlgorithmHyperparameter'
    ModelHyperparameter = 'ModelHyperparameter'
    SamplingHyperparameter = 'SamplingHyperparameter'
    MetricHyperparameter = 'MetricHyperparameter'


@dataclass
class PredefinedAIStringArg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    value: str
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.String

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value


@dataclass
class PredefinedAIBoolArg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    value: bool
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.Bool

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value


@dataclass
class PredefinedAIIntegerArg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    min: int
    max: int
    value: int
    unit_step: int
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.Integer

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value


@dataclass
class PredefinedAIFloatArg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    min: float
    max: float
    value: float
    unit_step: float
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.Float

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value


@dataclass
class PredefinedAIMultipleOf2Arg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    min: int
    max: int
    value: int
    unit_step: int
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.MultipleOf2

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value


@dataclass
class PredefinedAICategoricalValue:
    key: str
    display_name: str
    description: str


@dataclass
class PredefinedAICategoricalArg(PredefinedAIArg):
    key: str
    display_name: str
    description: str
    values: List[PredefinedAICategoricalValue]
    value: PredefinedAICategoricalValue
    group: ArgGroup = ArgGroup.Arg
    type: PredefinedAIArgType = PredefinedAIArgType.Categorical

    def get_key(self):
        return self.key

    def get_value(self):
        return self.value.key


@dataclass
class ModelConfig:
    key: str
    display_name: str
    description: str
    hyperparameters: List[
        Union[PredefinedAIBoolArg, PredefinedAIIntegerArg, PredefinedAIFloatArg, PredefinedAIMultipleOf2Arg, PredefinedAICategoricalArg]]

    def __post_init__(self):
        self.hyperparameter_dict = {}
        for hp in self.__getattribute__('hyperparameters'):
            self.hyperparameter_dict[hp.get_key()] = hp.get_value()


@dataclass
class StageConfig:
    args: List[
        Union[
            PredefinedAIStringArg, PredefinedAIBoolArg, PredefinedAIIntegerArg, PredefinedAIFloatArg, PredefinedAIMultipleOf2Arg, PredefinedAICategoricalArg]]


@dataclass
class PredefinedAIModelConfig:
    display_name: str
    key: str
    description: str
    model_configs: List[ModelConfig]
    train: StageConfig
    retrain: StageConfig
    inference: StageConfig
    serving: StageConfig

    def __post_init__(self):
        self.hyperparameter_dict = {}
        for model_config in self.model_configs:
            self.hyperparameter_dict[model_config.key] = {}
            for hp in model_config.hyperparameters:
                self.hyperparameter_dict[model_config.key][hp.get_key()] = hp.get_value()

        self.args_dict = {}
        for stage in ['train', 'retrain', 'inference', 'serving']:
            self.args_dict[stage] = {}
            for arg in self.__getattribute__(stage).__getattribute__('args'):
                self.args_dict[stage][arg.get_key()] = arg.get_value()

    def _get_model_hyperparameters_dict(self, model_name) -> dict:
        if model_name in self.hyperparameter_dict.keys():
            return self.hyperparameter_dict[model_name]
        raise ValueError(f'model "{model_name}" dose not exist in model_configs')

    def _get_args_dict(self, stage) -> dict:
        if stage in ['train', 'retrain', 'inference', 'serving']:
            return self.args_dict[stage]
        raise ValueError(f'stage "{stage}" dose not exist')



from dataclasses import dataclass
from enum import Enum
from typing import List, Union
from typing_extensions import TypedDict


class StrEnum(str, Enum):
    def __repr__(self):
        return str(self.name)

    def __str__(self):
        return str(self.name)


class ColumnType(StrEnum):
    Numerical = 'Numerical'
    Categorical = 'Categorical'
    Textual = 'Textual'
    TimeStamp = 'TimeStamp'
    Auto = 'Auto'
    Empty = 'Empty'


@dataclass
class Column:
    name: str
    type: ColumnType


class ColumnDict(TypedDict):
    name: str
    type: ColumnType


class DataConfigurationType(StrEnum):
    Data = 'Data'
    Column = 'Column'
    Columns = 'Columns'


# class DefinedConstraint(StrEnum):
#     ColumnNumberConstraint = 'ColumnNumberConstraint'
#     ColumnRequiredConstraint = 'ColumnRequiredConstraint'
#     RecordNumberConstraint = 'RecordNumberConstraint'
#
#
# @dataclass
# class Constraint:
#     dataKey: str
#     key: str
#     type: DataConfigurationType
#     constraint: DefinedConstraint
#     value: int


@dataclass
class DataConfiguration:
    display_name: str
    key: str
    type: DataConfigurationType
    value: Union[Column, List[Column]]
    required: bool
    # constraints: List[Constraint]
    description: str

    def __post_init__(self):
        if self.value is None:
            if self.type == DataConfigurationType.Column:
                self.value = Column(name='', type=ColumnType.Empty)
            elif self.type == DataConfigurationType.Columns:
                self.value = []


@dataclass
class PredefinedAIDataConfig:
    display_name: str
    key: str
    description: str
    configurations: List[DataConfiguration]
    # constraints: List[Constraint]




from mlplatform_lib.predefinedai.predefinedai_data_config import PredefinedAIDataConfig, DataConfigurationType, ColumnDict, Column
from mlplatform_lib.predefinedai.predefinedai_model_config import PredefinedAIModelConfig
import yaml
import argparse
from typing import List, Union, Dict
from inspect import signature
from inflection import underscore, camelize
from dataclasses import is_dataclass
from enum import EnumMeta
import json


class PredefinedAIConfigParser:
    def __init__(self, data_config_path=None, model_config_path=None):
        parser = argparse.ArgumentParser()
        parser.add_argument("--data_config", type=str, default=None)
        parser.add_argument("--model_config", type=str, default=None)
        _args = parser.parse_args()

        if _args.data_config is not None:
            data_config_dict = json.loads(_args.data_config)
        elif data_config_path is not None:
            with open(data_config_path) as data_config_file:
                data_config_dict = yaml.safe_load(data_config_file)
        else:
            raise ValueError('data_config must be in args')

        if _args.model_config is not None:
            model_config_dict = json.loads(_args.model_config)
        elif model_config_path is not None:
            with open(model_config_path) as model_config_file:
                model_config_dict = yaml.safe_load(model_config_file)
        else:
            raise ValueError('model_config must be in args')

        self.data_config_list: List[PredefinedAIDataConfig] = [from_dict(PredefinedAIDataConfig, data_config_d) for
                                                               data_config_d in data_config_dict]
        self.model_config: PredefinedAIModelConfig = from_dict(PredefinedAIModelConfig, model_config_dict)

    def get_data_config_from_data_key(self, data_key) -> PredefinedAIDataConfig:
        try:
            return list(filter(lambda x: x.__getattribute__('key') == data_key, self.data_config_list))[0]
        except IndexError:
            raise ValueError(f'data_key "{data_key}" do not exist in data_config')

    def get_data_config_dict_from_data_key(self, data_key) -> Dict[str, Union[ColumnDict, List[ColumnDict]]]:
        data_config_dict = {}
        data_config = self.get_data_config_from_data_key(data_key)
        for config in data_config.configurations:
            if config.type == DataConfigurationType.Column:
                if config.value is None:
                    data_config_dict[config.key] = None
                else:
                    data_config_dict[config.key] = {'name': config.value.name, 'type': config.value.type}
            elif config.type == DataConfigurationType.Columns:
                data_config_dict[config.key] = []
                if config.value is None:
                    continue
                for value in config.value:
                    data_config_dict[config.key].append({'name': value.name, 'type': value.type})
        return data_config_dict

    def get_all_data_config_dict(self) -> Dict[str, Dict[str, Union[ColumnDict, List[ColumnDict]]]]:
        all_data_config_dict = {}
        data_keys = [data_config.key for data_config in self.data_config_list]
        for data_key in data_keys:
            all_data_config_dict[data_key] = self.get_data_config_dict_from_data_key(data_key)
        return all_data_config_dict

    def get_model_name(self) -> str:
        return self.model_config.key

    def get_sub_model_name_list(self) -> List[str]:
        return [mc.key for mc in self.model_config.model_configs]

    def get_model_hyperparameters_dict(self, model_name: str) -> dict:
        return self.model_config._get_model_hyperparameters_dict(model_name)

    def get_train_args_dict(self) -> dict:
        return self._get_stage_args_dict('train')

    def get_retrain_ars_dict(self) -> dict:
        return self._get_stage_args_dict('retrain')

    def get_inference_args_dict(self) -> dict:
        return self._get_stage_args_dict('inference')

    def get_serving_args_dict(self) -> dict:
        return self._get_stage_args_dict('serving')

    def _get_stage_args_dict(self, stage) -> dict:
        return self.model_config._get_args_dict(stage)


def from_dict(dataclass, dictionary):
    if dictionary is None:
        return None
    temp_dict = {}
    sig = signature(dataclass)
    for key, value in dictionary.items():
        key = underscore(key)
        assert key in sig.parameters.keys(), \
            f'key ({key}) dose not exist in {dataclass.__name__} keys ({dataclass.__annotations__.keys()}), ' \
            f'({dataclass.__name__}: {dictionary})'
        key_class = sig.parameters[key].annotation
        if is_dataclass(key_class):
            temp_dict[key] = from_dict(key_class, value)
        elif (
            hasattr(key_class, '__origin__') and
            key_class.__origin__ == list
        ):
            # temp_dict[key] = [from_dict(key_class.__args__[0], v) for v in value]
            if is_dataclass(key_class.__args__[0]):
                temp_dict[key] = [
                    from_dict(key_class.__args__[0], v) for v in value
                ]
            elif (
                hasattr(key_class.__args__[0], '__origin__') and
                key_class.__args__[0].__origin__ == Union
            ):
                def union_mapping(union_classes, value):
                    for union_class in union_classes.__args__:
                        if value is None or set(union_class.__annotations__.keys()) == {underscore(v_k) for v_k in value.keys()}:
                            try:
                                return from_dict(union_class, value)
                            except AssertionError:
                                continue
                    raise ValueError(f'value ({value}) dose not fit in dataclasses ({union_classes.__args__})')
                temp_dict[key] = [union_mapping(key_class.__args__[0], v) for v in value]
        elif (
            hasattr(key_class, '__origin__') and key_class.__origin__ == Union
        ):
            def union_mapping(union_classes, value):
                for union_class in union_classes.__args__:
                    if value is None or set(union_class.__annotations__.keys()) == {underscore(v_k) for v_k in value.keys()}:
                        try:
                            return from_dict(union_class, value)
                        except AssertionError:
                            continue
                raise ValueError(f'value ({value}) dose not fit in dataclasses ({union_classes.__args__})')
            if type(value) == list:
                temp_dict[key] = [union_mapping(key_class, v) for v in value]
            else:
                temp_dict[key] = union_mapping(key_class, value)
        elif isinstance(key_class, EnumMeta):
            temp_dict[key] = key_class(value)
        else:
            assert isinstance(value, dataclass.__annotations__[key]), \
                    f'{key} type must be {dataclass.__annotations__[key]} (type: {type(value)}, value: {value}))'
            temp_dict[key] = value
    return dataclass(**temp_dict)
